"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const database_1 = require("@nocobase/database");
const json_templates_1 = __importDefault(require("json-templates"));
const constants_1 = require("../constants");
const instructions_1 = __importDefault(require("../instructions"));
const calculators_1 = __importDefault(require("../calculators"));
class ExecutionModel extends database_1.Model {
    constructor() {
        super(...arguments);
        this.nodes = [];
        this.nodesMap = new Map();
        this.jobsMap = new Map();
        this.jobsMapByNodeId = {};
    }
    // make dual linked nodes list then cache
    makeNodes(nodes = []) {
        this.nodes = nodes;
        nodes.forEach((node) => {
            this.nodesMap.set(node.id, node);
        });
        nodes.forEach((node) => {
            if (node.upstreamId) {
                node.upstream = this.nodesMap.get(node.upstreamId);
            }
            if (node.downstreamId) {
                node.downstream = this.nodesMap.get(node.downstreamId);
            }
        });
    }
    makeJobs(jobs) {
        jobs.forEach((job) => {
            this.jobsMap.set(job.id, job);
            // TODO: should consider cycle, and from previous job
            this.jobsMapByNodeId[job.nodeId] = job.result;
        });
    }
    getTransaction() {
        return __awaiter(this, void 0, void 0, function* () {
            const { sequelize } = this.constructor.database;
            if (!this.useTransaction) {
                return undefined;
            }
            const { options } = this;
            // @ts-ignore
            const transaction = options.transaction && !options.transaction.finished
                ? options.transaction
                : sequelize.transaction();
            // @ts-ignore
            if (this.transaction !== transaction.id) {
                // @ts-ignore
                yield this.update({ transaction: transaction.id }, { transaction });
            }
            return transaction;
        });
    }
    prepare(options, commit = false) {
        return __awaiter(this, void 0, void 0, function* () {
            this.options = options || {};
            const transaction = yield this.getTransaction();
            this.transaction = transaction;
            if (!this.workflow) {
                this.workflow = yield this.getWorkflow({ transaction });
            }
            const nodes = yield this.workflow.getNodes({ transaction });
            this.makeNodes(nodes);
            const jobs = yield this.getJobs({
                order: [['id', 'ASC']],
                transaction,
            });
            this.makeJobs(jobs);
            if (commit) {
                yield this.commit();
            }
        });
    }
    start(options) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.status !== constants_1.EXECUTION_STATUS.STARTED) {
                throw new Error(`execution was ended with status ${this.status}`);
            }
            yield this.prepare(options);
            if (this.nodes.length) {
                const head = this.nodes.find(item => !item.upstream);
                yield this.run(head, { result: this.context });
            }
            else {
                yield this.exit(null);
            }
            yield this.commit();
        });
    }
    resume(job, options) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.status !== constants_1.EXECUTION_STATUS.STARTED) {
                throw new Error(`execution was ended with status ${this.status}`);
            }
            yield this.prepare(options);
            const node = this.nodesMap.get(job.nodeId);
            yield this.recall(node, job);
            yield this.commit();
        });
    }
    commit() {
        return __awaiter(this, void 0, void 0, function* () {
            // @ts-ignore
            if (this.transaction && (!this.options.transaction || this.options.transaction.finished)) {
                yield this.transaction.commit();
            }
        });
    }
    exec(instruction, node, prevJob) {
        return __awaiter(this, void 0, void 0, function* () {
            let job;
            try {
                // call instruction to get result and status
                job = yield instruction.call(node, prevJob, this);
                if (!job) {
                    return null;
                }
            }
            catch (err) {
                // for uncaught error, set to rejected
                job = {
                    result: err instanceof Error
                        ? { message: err.message, stack: process.env.NODE_ENV === 'production' ? [] : err.stack }
                        : err,
                    status: constants_1.JOB_STATUS.REJECTED,
                };
                // if previous job is from resuming
                if (prevJob && prevJob.nodeId === node.id) {
                    prevJob.set(job);
                    job = prevJob;
                }
            }
            let savedJob;
            // TODO(optimize): many checking of resuming or new could be improved
            // could be implemented separately in exec() / resume()
            if (job instanceof database_1.Model) {
                savedJob = (yield job.save({ transaction: this.transaction }));
            }
            else {
                const upstreamId = prevJob instanceof database_1.Model ? prevJob.get('id') : null;
                savedJob = yield this.saveJob(Object.assign({ nodeId: node.id, upstreamId }, job));
            }
            if (savedJob.status === constants_1.JOB_STATUS.RESOLVED && node.downstream) {
                // run next node
                return this.run(node.downstream, savedJob);
            }
            // all nodes in scope have been executed
            return this.end(node, savedJob);
        });
    }
    run(node, input) {
        return __awaiter(this, void 0, void 0, function* () {
            const { run } = instructions_1.default.get(node.type);
            if (typeof run !== 'function') {
                return Promise.reject(new Error('`run` should be implemented for customized execution of the node'));
            }
            return this.exec(run, node, input);
        });
    }
    // parent node should take over the control
    end(node, job) {
        const parentNode = this.findBranchParentNode(node);
        // no parent, means on main flow
        if (parentNode) {
            return this.recall(parentNode, job);
        }
        // really done for all nodes
        // * should mark execution as done with last job status
        return this.exit(job);
    }
    recall(node, job) {
        return __awaiter(this, void 0, void 0, function* () {
            const { resume } = instructions_1.default.get(node.type);
            if (typeof resume !== 'function') {
                return Promise.reject(new Error('`resume` should be implemented because the node made branch'));
            }
            return this.exec(resume, node, job);
        });
    }
    exit(job) {
        return __awaiter(this, void 0, void 0, function* () {
            const status = job ? ExecutionModel.StatusMap[job.status] : constants_1.EXECUTION_STATUS.RESOLVED;
            yield this.update({ status }, { transaction: this.transaction });
            return null;
        });
    }
    // TODO(optimize)
    saveJob(payload) {
        return __awaiter(this, void 0, void 0, function* () {
            const { database } = this.constructor;
            const { model } = database.getCollection('jobs');
            const [job] = (yield model.upsert(Object.assign(Object.assign({}, payload), { executionId: this.id }), { transaction: this.transaction }));
            this.jobsMap.set(job.id, job);
            this.jobsMapByNodeId[job.nodeId] = job.result;
            return job;
        });
    }
    // find the first node in current branch
    findBranchStartNode(node) {
        for (let n = node; n; n = n.upstream) {
            if (n.branchIndex !== null) {
                return n;
            }
        }
        return null;
    }
    // find the node start current branch
    findBranchParentNode(node) {
        for (let n = node; n; n = n.upstream) {
            if (n.branchIndex !== null) {
                return n.upstream;
            }
        }
        return null;
    }
    findBranchParentJob(job, node) {
        for (let j = job; j; j = this.jobsMap.get(j.upstreamId)) {
            if (j.nodeId === node.id) {
                return j;
            }
        }
        return null;
    }
    getParsedValue(value, node) {
        const injectedFns = {};
        const scope = {
            execution: this,
            node
        };
        for (let [name, fn] of calculators_1.default.getEntities()) {
            injectedFns[name] = fn.bind(scope);
        }
        return (0, json_templates_1.default)(value)({
            $context: this.context,
            $jobsMapByNodeId: this.jobsMapByNodeId,
            $fn: injectedFns
        });
    }
}
exports.default = ExecutionModel;
ExecutionModel.StatusMap = {
    [constants_1.JOB_STATUS.PENDING]: constants_1.EXECUTION_STATUS.STARTED,
    [constants_1.JOB_STATUS.RESOLVED]: constants_1.EXECUTION_STATUS.RESOLVED,
    [constants_1.JOB_STATUS.REJECTED]: constants_1.EXECUTION_STATUS.REJECTED,
    [constants_1.JOB_STATUS.CANCELLED]: constants_1.EXECUTION_STATUS.CANCELLED,
};
//# sourceMappingURL=Execution.js.map